# GRU simple normal nn, this model assumes the two elements in state space is uncorrelated.
import tensorflow as tf
import numpy as np
from collections import deque
from utils import Grid
from itertools import product
import pandas as pd
from copy import deepcopy



DEBUG = False


class FC_Model_Deter:
    def __init__(self,
                 sess, 
                 state_dim, 
                 action_dim, 
                 steps = None, 
                 pool_maxlen = 10000,
                 learning_rate = 0.01,
                 *args, **kwargs):
        self.sess = sess
        self.state_dim, self.action_dim, self.steps = state_dim, action_dim, steps
        self.std_c = 1. # coefficient for variance loss 
        self.lr = learning_rate
        self.inputs = {}
        self.outputs = {}
        self.other_tensors = {}
        self.build()
        self.tensors = {}
        
        self.tensors.update(self.inputs)
        self.tensors.update(self.outputs)
        self.tensors.update(self.other_tensors)
        init = tf.global_variables_initializer()
        # self.sess.run(init)
        self.is_trained = False
        self.data_pool = {"states": deque(maxlen = pool_maxlen), 
                          "actions": deque(maxlen = pool_maxlen), 
                          "next_states": deque(maxlen = pool_maxlen), 
                          "counter": 0}
        self.pool_maxlen = pool_maxlen                
        
    
    def build(self, hidden_size = (64, 64)):
            
        action = tf.keras.Input(shape=(self.action_dim,), name = "action_ipt")
        state = tf.keras.Input(shape=(self.state_dim,), name = "state_ipt")
        self.inputs["action"] = action
        self.inputs["state"] = state

        x = tf.keras.layers.Concatenate(axis = 1)([state, action])
        # 
        if DEBUG: print(x)
        for i,size in enumerate(hidden_size):
            
            x =  tf.keras.layers.Dense(size, name = f"fc_layer_{i}", activation = "relu")(x)
        
        # mean
        out = tf.keras.layers.Dense(self.state_dim, name = "mean")(x)
        
        self.outputs["output"] = out
        # print(self.inputs.values)
        self.model = tf.keras.Model(inputs = list(self.inputs.values()), outputs = list(self.outputs.values()))
        print(self.model.summary())

        
        true_y = tf.keras.Input(shape=(self.state_dim, ))
        

        loss = tf.losses.mean_squared_error(labels = true_y, predictions=out)

        updt = tf.train.AdamOptimizer(self.lr).minimize(loss)
        
        self.other_tensors["loss"] = loss
        self.other_tensors["loss"] = loss
        self.other_tensors["true_y"] = true_y
        self.other_tensors["updt"] = updt
       
    def get_tensors(self, keys):
        return [self.tensors[key] for key in keys]
    
    def train(self, iters = 10000):
        data = self.arrange_data()
        n_update = 10
        n_iter = int(iters/n_update)
        for _ in range(n_iter):
            self.do_training(data['train_s'], data['train_a'], data['train_es'], n_update)
            val_loss = self.validate(data["test_s"], data['test_a'], data['test_es'])
            print(f"val_loss: {val_loss}")

    def do_training(self, train_s, train_a, train_es , train_iter):
        updt, loss, action, true_y, state = self.get_tensors(["updt", 'loss', 'action', 'true_y', 'state']) 
        
        for _ in range(train_iter):
            _, l_ = self.sess.run((updt, loss), feed_dict={action: train_a, true_y : train_es, state: train_s})
            if DEBUG: print(f'loss: {l_}')
        print(f'Training finished: {train_iter} updates.')
        self.is_trained = True
        return l_
    
    
    
    def validate(self, train_s, train_a, train_es):
        updt, loss, actions, true_y, init_state = self.get_tensors(["updt", 'loss', 'action', 'true_y', 'state']) 
        l_ = self.sess.run(loss, feed_dict={actions: train_a, true_y : train_es, init_state: train_s})
        return l_
    
    def plan(self, s, as_, n_traj):
        init_s = s.copy()
        if self.is_trained:
            ret = []
            for _ in range(n_traj):
                s = [init_s]
                each_traj = [s[0]]
                for a in as_:
                    m =  self.sess.run(self.tensors["output"], 
                                            feed_dict={self.tensors["action"]: [a], self.tensors["state"]: s})
                    
                    s = m
                    # if DEBUG: print("shape of mean and variance" ,m.shape, var.shape, rnds.shape, s.shape)
                    each_traj.append(s.reshape(-1))
                ret.append(each_traj)
            #print(m.shape, c.shape, rnds.shape)
            # print(ret)
            return np.array(ret, dtype=float)
        else:
            raise
    
    
    def run(self, input_keys, output_keys, feed_dict):
        return self.sess.run([self.tensors[k] for k in output_keys], feed_dict={self.tensors[k]: feed_dict[k] for k in input_keys})

    def add_data(self, state, action, next_state):
        self.data_pool["states"].append(state)
        self.data_pool["actions"].append(action)
        self.data_pool['next_states'].append(next_state)
        self.data_pool["counter"]  = min(self.data_pool["counter"]+1, self.pool_maxlen)
        

    def arrange_data(self, val_ratio = 0.2):
        states = self.data_pool['states']
        actions = self.data_pool['actions']
        next_states = self.data_pool["next_states"]
        n = self.data_pool['counter']
        
        shuffled_idx = np.random.choice(range(n-1), size = n-1, replace = False) 
        split_idx = int(val_ratio*(n-1))
        
        
        states = np.array(states).reshape(n, -1)
        actions = np.array(actions).reshape(n, -1)
        next_states = np.array(next_states).reshape(n, -1)

        train_s = states[shuffled_idx[0:split_idx]]
        train_a = actions[shuffled_idx[0:split_idx]]
        train_es = next_states[shuffled_idx[0:split_idx]]

        test_s = states[shuffled_idx[split_idx:]]
        test_a = actions[shuffled_idx[split_idx:]]
        test_es = next_states[shuffled_idx[split_idx:]]

        
        if DEBUG:
            print(train_es)

        return {"train_a": train_a,
                "test_a": test_a,
                "train_s": train_s,
                "test_s": test_s,
                "train_es": train_es,
                "test_es": test_es}    

class FC_Hist_Model:
    def __init__(self,
                 sess, 
                 state_dim, 
                 action_dim, 
                 output_grid,
                 steps = None, 
                 pool_maxlen = 10000,
                 learning_rate = 0.0001,
                 *args, **kwargs):
        self.sess = sess
        self.state_dim, self.action_dim, self.steps = state_dim, action_dim, steps
        self.output_grid = output_grid # should be a two-dim list shape: d*n, d state_dim, n discretize dim.
        self.output_map = {i: each_sub_grid_index
                            for i, each_sub_grid_index in 
                            enumerate(product(*[np.digitize(list(bins)+[bins[-1]+1], bins, right = True) 
                                     for bins in output_grid]))}
        
        self.output_inv_map = {tuple(v): k for k, v in self.output_map.items()}
        self.output_dim = len(self.output_map)
        self.lr = learning_rate
        self.inputs = {}
        self.outputs = {}
        self.other_tensors = {}
        self.build()
        self.tensors = {}
        
        self.tensors.update(self.inputs)
        self.tensors.update(self.outputs)
        self.tensors.update(self.other_tensors)
        init = tf.global_variables_initializer()
        # self.sess.run(init)
        self.is_trained = False
        self.data_pool = {"states": deque(maxlen = pool_maxlen), 
                          "actions": deque(maxlen = pool_maxlen), 
                          "next_states": deque(maxlen = pool_maxlen),
                          "counter": 0}
        self.pool_maxlen = pool_maxlen                
    
    def multi_dim_digitize(self, xs):
        grid = self.output_grid
        if len(xs.shape) == 1:
            xs = np.array([xs])
            
        ret = [self.output_inv_map[pair] for pair in 
            zip(*[np.digitize(xs[:, i], grid[i]).tolist() 
                    for i in range(len(grid))])]
        if len(xs.shape) == 1:
            return ret[0]
        else:
            return np.array(ret, dtype = int)
        
    def _one_hot(self, xs):
        if len(xs.shape) == 1:
            xs = np.array([xs])
        
        batch_size = xs.shape[0]
        indices = self.multi_dim_digitize(xs)
        
        ret = np.zeros(shape = (batch_size, self.output_dim))
        ret[list(range(batch_size)), indices] = 1
        if len(xs.shape) == 1:
            return ret[0]
        else:
            return np.array(ret, dtype = int)
            
    def build(self, hidden_size = (64, 64)):
            
        action = tf.keras.Input(shape=(self.action_dim,), name = "action_ipt")
        state = tf.keras.Input(shape=(self.state_dim,), name = "state_ipt")
        self.inputs["action"] = action
        self.inputs["state"] = state

        x = tf.keras.layers.Concatenate(axis = 1)([state, action])
        # 
        if DEBUG: print(x)
        for i,size in enumerate(hidden_size):
            
            x =  tf.keras.layers.Dense(size, name = f"fc_layer_{i}", activation = "relu")(x)
        
        # mean
        out = tf.keras.layers.Dense(self.output_dim, activation = "softmax", name = "hist")(x)
        
        self.outputs["output"] = out
        # print(self.inputs.values)
        self.model = tf.keras.Model(inputs = list(self.inputs.values()), outputs = list(self.outputs.values()))
        print(self.model.summary())

        
        true_y = tf.keras.Input(shape=(self.output_dim, ))
        
        loss = tf.keras.losses.CategoricalCrossentropy(labels = true_y, predictions=out)

        updt = tf.train.AdamOptimizer(self.lr).minimize(loss)
        
        self.other_tensors["loss"] = loss
        self.other_tensors["loss"] = loss
        self.other_tensors["true_y"] = true_y
        self.other_tensors["updt"] = updt
       
    def get_tensors(self, keys):
        return [self.tensors[key] for key in keys]
    
    def train(self, iters = 10000):
        # TODO
        data = self.arrange_data()
        n_update = 10
        n_iter = int(iters/n_update)
        for _ in range(n_iter):
            self.do_training(data['train_s'], data['train_a'], data['train_es'], n_update)
            val_loss = self.validate(data["test_s"], data['test_a'], data['test_es'])
            print(f"val_loss: {val_loss}")

    def do_training(self, train_s, train_a, train_es , train_iter):
        # TODO
        updt, loss, action, true_y, state = self.get_tensors(["updt", 'loss', 'action', 'true_y', 'state']) 
        
        for _ in range(train_iter):
            _, l_ = self.sess.run((updt, loss), feed_dict={action: train_a, true_y : train_es, state: train_s})
            if DEBUG: print(f'loss: {l_}')
        print(f'Training finished: {train_iter} updates.')
        self.is_trained = True
        return l_
    
    
    
    def validate(self, train_s, train_a, train_es):
        # TODO
        updt, loss, actions, true_y, init_state = self.get_tensors(["updt", 'loss', 'action', 'true_y', 'state']) 
        l_ = self.sess.run(loss, feed_dict={actions: train_a, true_y : train_es, init_state: train_s})
        return l_
    
    def plan(self, s, as_, n_traj):
        # TODO
        init_s = s.copy()
        if self.is_trained:
            ret = []
            for _ in range(n_traj):
                s = [init_s]
                each_traj = [s[0]]
                for a in as_:
                    m =  self.sess.run(self.tensors["output"], 
                                            feed_dict={self.tensors["action"]: [a], self.tensors["state"]: s})
                    
                    s = m
                    # if DEBUG: print("shape of mean and variance" ,m.shape, var.shape, rnds.shape, s.shape)
                    each_traj.append(s.reshape(-1))
                ret.append(each_traj)
            #print(m.shape, c.shape, rnds.shape)
            # print(ret)
            return np.array(ret, dtype=float)
        else:
            raise
    
    
    def run(self, input_keys, output_keys, feed_dict):
        return self.sess.run([self.tensors[k] for k in output_keys], feed_dict={self.tensors[k]: feed_dict[k] for k in input_keys})

    def add_data(self, state, action, next_state):
        self.data_pool["states"].append(state)
        self.data_pool["actions"].append(action)
        self.data_pool['next_states'].append(next_state)
        self.data_pool["counter"]  = min(self.data_pool["counter"]+1, self.pool_maxlen)
        

    def arrange_data(self, val_ratio = 0.2):
        states = self.data_pool['states']
        actions = self.data_pool['actions']
        next_states = self.data_pool["next_states"]
        n = self.data_pool['counter']
        
        shuffled_idx = np.random.choice(range(n-1), size = n-1, replace = False) 
        split_idx = int(val_ratio*(n-1))
        
        
        states = np.array(states).reshape(n, -1)
        actions = np.array(actions).reshape(n, -1)
        next_states = np.array(next_states).reshape(n, -1)

        train_s = states[shuffled_idx[0:split_idx]]
        train_a = actions[shuffled_idx[0:split_idx]]
        train_es = next_states[shuffled_idx[0:split_idx]]

        test_s = states[shuffled_idx[split_idx:]]
        test_a = actions[shuffled_idx[split_idx:]]
        test_es = next_states[shuffled_idx[split_idx:]]

        
        if DEBUG:
            print(train_es)

        return {"train_a": train_a,
                "test_a": test_a,
                "train_s": train_s,
                "test_s": test_s,
                "train_es": train_es,
                "test_es": test_es}    


class FC_Model:
    def __init__(self,
                 sess, 
                 state_dim, 
                 action_dim, 
                 steps = None, 
                 pool_maxlen = 10000,
                 learning_rate = 5e-5,
                 *args, **kwargs):
        self.sess = sess
        self.state_dim, self.action_dim, self.steps = state_dim, action_dim, steps
        self.std_c = 1. # coefficient for variance loss 
        self.lr = learning_rate
        self.inputs = {}
        self.outputs = {}
        self.other_tensors = {}
        self.build()
        self.is_trained = True
        self.tensors = {}
        
        self.tensors.update(self.inputs)
        self.tensors.update(self.outputs)
        self.tensors.update(self.other_tensors)
        init = tf.global_variables_initializer()
        # self.sess.run(init)

        self.data_pool = {"states": deque(maxlen = pool_maxlen), 
                          "actions": deque(maxlen = pool_maxlen), 
                          "next_states": deque(maxlen = pool_maxlen),
                          "counter": 0}
        self.pool_maxlen = pool_maxlen                
        
    
    def build(self, hidden_size = (64, 64)):
            
        action = tf.keras.Input(shape=(self.action_dim,), name = "action_ipt")
        state = tf.keras.Input(shape=(self.state_dim,), name = "state_ipt")
        self.inputs["action"] = action
        self.inputs["state"] = state

        # action_one_hot = tf.one_hot(tf.reshape(action, [-1]), )
        x = tf.keras.layers.Concatenate(axis = 1)([state, action])
        x1, x2 = x, x
        # 
        if DEBUG: print(x)
        for i,size in enumerate(hidden_size):
            
            x1 =  tf.keras.layers.Dense(size, name = f"fc_layer_mean_{i}", activation = "relu")(x1)
            x2 =  tf.keras.layers.Dense(size, name = f"fc_layer_var_{i}", activation = "relu")(x2)
            

        
        # mean
        mean = tf.keras.layers.Dense(self.state_dim, name = "mean")(x1)
        
        # var
        logvar = tf.keras.layers.Dense(self.state_dim, name = "logvar")(x2)

        self.outputs["mean"] = mean
        self.outputs['logvar'] = logvar
        # print(self.inputs.values)
        self.model = tf.keras.Model(inputs = list(self.inputs.values()), outputs = list(self.outputs.values()))
        print(self.model.summary())


        true_y = tf.keras.Input(shape=(self.state_dim, ))

        inv_var = tf.exp(-logvar)

        mse_losses = tf.reduce_mean(tf.reduce_mean(tf.square(mean - true_y) * inv_var, axis=-1), axis=-1)
        var_losses = tf.reduce_mean(tf.reduce_mean(logvar, axis=-1), axis=-1)
        total_losses = mse_losses + var_losses
        
        
        target_var = (true_y-mean)**2
        # mean_loss = tf.losses.mean_squared_error(labels = true_y, predictions=mean)
        # var_loss = tf.losses.mean_squared_error(labels = target_var, predictions=var)
        
        # loss = mean_loss+ self.std_c*var_loss
        updt = tf.train.AdamOptimizer(self.lr).minimize(total_losses)
        
        self.other_tensors["target_var"] = target_var
        self.other_tensors["mean_loss"] = mse_losses
        self.other_tensors['var_loss'] = var_losses
        self.other_tensors["loss"] = total_losses
        self.other_tensors["true_y"] = true_y
        self.other_tensors["updt"] = updt
       
    def get_tensors(self, keys):
        return [self.tensors[key] for key in keys]
    
    def train(self, iters = 10000):
        data = self.arrange_data()
        n_update = 10
        n_iter = int(iters/n_update)
        for _ in range(n_iter):
            self.do_training(data['train_s'], data['train_a'], data['train_es'], n_update)
            val_loss = self.validate(data["test_s"], data['test_a'], data['test_es'])
            # print(f"val_loss: {val_loss}")

    def do_training(self, train_s, train_a, train_es , train_iter):
        updt, loss, action, true_y, state, mean_loss, var_loss = self.get_tensors(["updt", 'loss', 'action', 'true_y', 'state', 'mean_loss', "var_loss"]) 
        
        for _ in range(train_iter):
            _, l_, ml, vl = self.sess.run((updt, loss, mean_loss, var_loss), feed_dict={action: train_a, true_y : train_es, state: train_s})
            if DEBUG: print(f'loss: {l_}={ml}+{vl}')
        # print(f'Training finished: {train_iter} updates.')
        self.is_trained = True
        return l_
    
    def validate(self, train_s, train_a, train_es):
        updt, loss, actions, true_y, init_state = self.get_tensors(["updt", 'loss', 'action', 'true_y', 'state']) 
        l_ = self.sess.run(loss, feed_dict={actions: train_a, true_y : train_es, init_state: train_s})
        return l_
    
    def plan(self, s, as_, n_traj, det = False):
        init_s = s.copy()
        if self.is_trained:
            ret = np.zeros(shape = (n_traj, len(as_)+1, len(s)))
            ret[:, 0, :] = init_s
            for i, a in enumerate(as_):
                m, logvar =  self.sess.run((self.tensors["mean"], self.tensors["logvar"]), 
                                        feed_dict={self.tensors["action"]: np.ones(shape = (n_traj, 1))*a, self.tensors["state"]: ret[:, i, :]})
                c = np.sqrt(np.exp(logvar))
                rnds = np.random.normal(size = (n_traj, self.state_dim))
                s = m+rnds*c*(0. if det else 1.)
                if DEBUG: print("shape of mean and variance" ,m.shape, logvar.shape, rnds.shape, s.shape)
                if DEBUG: print(f'variance: {np.exp(logvar)}')
                
                ret[:, i+1, :] = s

            #print(m.shape, c.shape, rnds.shape)
            # print(ret)
            return np.array(ret, dtype=float) # shape: (n_traj, delay_steps, state_dim)
        else:
            raise NotImplementedError("Model isn't trained.")
    
    
    def run(self, input_keys, output_keys, feed_dict):
        return self.sess.run([self.tensors[k] for k in output_keys], feed_dict={self.tensors[k]: feed_dict[k] for k in input_keys})

    def add_data(self, state, action, next_state):
        self.data_pool["states"].append(state)
        self.data_pool["actions"].append(action)
        self.data_pool['next_states'].append(next_state)
        self.data_pool["counter"]  = min(self.data_pool["counter"]+1, self.pool_maxlen)
        

    def arrange_data(self, val_ratio = 0.2):
        states = self.data_pool['states']
        actions = self.data_pool['actions']
        next_states = self.data_pool['next_states']
        n = self.data_pool['counter']
        
        shuffled_idx = np.random.choice(range(n-1), size = n-1, replace = False) 
        split_idx = int(val_ratio*(n-1))
        
        
        states = np.array(states).reshape(n, -1)
        actions = np.array(actions).reshape(n, -1)
        next_states = np.array(next_states).reshape(n, -1)

        train_s = states[shuffled_idx[0:split_idx]]
        train_a = actions[shuffled_idx[0:split_idx]]
        train_es = next_states[shuffled_idx[0:split_idx]]

        test_s = states[shuffled_idx[split_idx:]]
        test_a = actions[shuffled_idx[split_idx:]]
        test_es = next_states[shuffled_idx[split_idx:]]

        
        if DEBUG:
            print(train_es)

        return {"train_a": train_a,
                "test_a": test_a,
                "train_s": train_s,
                "test_s": test_s,
                "train_es": train_es,
                "test_es": test_es}
        
        
class FC_Model_Disc(FC_Model):
    def __init__(self,
                 sess, 
                 state_dim, 
                 action_num, 
                 steps = None, 
                 pool_maxlen = 10000,
                 learning_rate = 5e-5,
                 *args, **kwargs):
        
        self.action_num  = action_num  
        FC_Model.__init__(self, sess, state_dim, 1, steps, pool_maxlen, learning_rate, *args, **kwargs)
                   
        
    
    def build(self, hidden_size = (64, 64)):
            
        action = tf.keras.Input(shape=(self.action_dim,), name = "action_ipt", dtype = tf.int32)
        state = tf.keras.Input(shape=(self.state_dim,), name = "state_ipt")
        self.inputs["action"] = action
        self.inputs["state"] = state

        action_one_hot = tf.one_hot(tf.reshape(action, [-1]), self.action_num)
        x = tf.keras.layers.Concatenate(axis = 1)([state, action_one_hot])
        x1, x2 = x, x
        # 
        if DEBUG: print(x)
        for i,size in enumerate(hidden_size):
            
            x1 =  tf.keras.layers.Dense(size, name = f"fc_layer_mean_{i}", activation = "relu")(x1)
            x2 =  tf.keras.layers.Dense(size, name = f"fc_layer_var_{i}", activation = "relu")(x2)
            

        
        # mean
        mean = tf.keras.layers.Dense(self.state_dim, name = "mean")(x1)
        
        # var
        logvar = tf.keras.layers.Dense(self.state_dim, name = "logvar")(x2)

        self.outputs["mean"] = mean
        self.outputs['logvar'] = logvar
        # print(self.inputs.values)
        self.model = tf.keras.Model(inputs = list(self.inputs.values()), outputs = list(self.outputs.values()))
        print(self.model.summary())


        true_y = tf.keras.Input(shape=(self.state_dim, ))

        inv_var = tf.exp(-logvar)

        mse_losses = tf.reduce_mean(tf.reduce_mean(tf.square(mean - true_y) * inv_var, axis=-1), axis=-1)
        var_losses = tf.reduce_mean(tf.reduce_mean(logvar, axis=-1), axis=-1)
        total_losses = mse_losses + var_losses
        
        
        target_var = (true_y-mean)**2
        # mean_loss = tf.losses.mean_squared_error(labels = true_y, predictions=mean)
        # var_loss = tf.losses.mean_squared_error(labels = target_var, predictions=var)
        
        # loss = mean_loss+ self.std_c*var_loss
        updt = tf.train.AdamOptimizer(self.lr).minimize(total_losses)
        
        self.other_tensors["target_var"] = target_var
        self.other_tensors["mean_loss"] = mse_losses
        self.other_tensors['var_loss'] = var_losses
        self.other_tensors["loss"] = total_losses
        self.other_tensors["true_y"] = true_y
        self.other_tensors["updt"] = updt

class Hist_Model_1D:
    def __init__(self, 
                 action_dim, 
                 action_space,
                 state_space_grid,
                 steps = None, 
                 pool_maxlen = 10000,
                 *args, **kwargs):
        self.state_dim, self.action_dim, self.action_space, self.steps = \
                            1, action_dim, action_space, steps
        self.state_space_grid = state_space_grid # should be one-dim ascending numpy array
        tmp = list(state_space_grid)
        tmp = [tmp[0]]+tmp+[tmp[-1]]
        self.index2interval_map = {i: (tmp[i], tmp[i+1])
                            for i, _ in 
                            enumerate(tmp[:-1])}

        self.state_space = list(self.index2interval_map.keys())
        self.is_trained = True
        self.frequency_matrix = {act: np.ones(shape = (len(self.index2interval_map), len(self.index2interval_map))) for act in action_space}
        self.trans_matrix = {act: np.zeros(shape = (len(self.index2interval_map), len(self.index2interval_map))) for act in action_space}
    
    def add_data(self, state, action, next_state):
        state, next_state = state[0], next_state[0]
        state, next_state = self.digitize([state, next_state])
        self.frequency_matrix[action][state, next_state] += 1
    
    def train(self, iters = 10000):
        for act in self.action_space:
            tmp_matrix = self.frequency_matrix[act]
            self.trans_matrix[act] = (tmp_matrix/tmp_matrix.sum(axis = 1).reshape(-1,1)).copy()
        self.is_trained = True
    
    def digitize(self, s):
        return np.digitize(s, self.state_space_grid)
    
    def plan(self, s, as_, n_traj, det = False):
        init_s = s.copy()
        if self.is_trained:
            ret = []
            for _ in range(n_traj):
                s = init_s
                each_traj = [s]
                for a in as_:
                    try:
                        # print(a)
                        # a = a[0]
                        pass
                    except TypeError:
                        pass
#                     print(f"s should be 1d array: {s}")
                    p = self.trans_matrix[a][self.digitize(s)[0], :]
                    # print(p, sum(p))
                    # print(s, a)
                    # print(p.argmax(), self.state_space[p.argmax()], self.index2interval_map[self.state_space[p.argmax()]])
                    if det:
                        s = self.state_space[p.argmax()]
                    else:
                        try:
                            s = np.random.choice(self.state_space, p = p)
                        except ValueError:
                            s = np.random.choice(self.state_space)
                    u, l = self.index2interval_map[s]
                    s = np.random.uniform(u, l)
                    s = np.array([s])
                    # print(s)
                    each_traj.append(s.reshape(-1))
                ret.append(each_traj)
            #print(m.shape, c.shape, rnds.shape)
            # print(ret)
            return np.array(ret, dtype=float)
        else:
            raise NotImplementedError("Model isn't trained.")
        
        
class Hist_Model_1D_Deter:
    def __init__(self, 
                 action_dim, 
                 action_space,
                 state_space_grid,
                 steps = None, 
                 pool_maxlen = 10000,
                 *args, **kwargs):
        self.state_dim, self.action_dim, self.action_space, self.steps = \
                            1, action_dim, action_space, steps
        self.state_space_grid = state_space_grid # should be one-dim ascending numpy array
        tmp = list(state_space_grid)
        tmp = [tmp[0]]+tmp+[tmp[-1]]
        self.index2interval_map = {i: (tmp[i], tmp[i+1])
                            for i, _ in 
                            enumerate(tmp[:-1])}

        self.state_space = list(self.index2interval_map.keys())
        self.is_trained = False
        self.frequency_matrix = {act: np.ones(shape = (len(self.index2interval_map), len(self.index2interval_map))) for act in action_space}
        self.trans_matrix = {act: np.zeros(shape = (len(self.index2interval_map), len(self.index2interval_map))) for act in action_space}
    
    def add_data(self, state, action, next_state):
        state, next_state = state[0], next_state[0]
        state, next_state = self.digitize([state, next_state])
        self.frequency_matrix[action][state, next_state] += 1
    
    def train(self, iters = 10000):
        for act in self.action_space:
            tmp_matrix = self.frequency_matrix[act]
            self.trans_matrix[act] = (tmp_matrix/tmp_matrix.sum(axis = 1).reshape(-1,1)).copy()
        self.is_trained = True
    
    def digitize(self, s):
        return np.digitize(s, self.state_space_grid)
    
    def plan(self, s, as_, n_traj):
        init_s = s.copy()
        if self.is_trained:
            ret = []
            for _ in range(n_traj):
                s = init_s
                each_traj = [s]
                for a in as_:
                    a = a[0]
#                     print(f"s should be 1d array: {s}")
                    p = self.trans_matrix[a][self.digitize(s)[0], :]
                    idx = np.array(p).reshape(-1).argmax()
                    s = self.state_space[idx]
                    u, l = self.index2interval_map[s]
                    s = (u+l)/2
                    s = np.array([s])
                    each_traj.append(s.reshape(-1))
                ret.append(each_traj)
            #print(m.shape, c.shape, rnds.shape)
            # print(ret)
            return np.array(ret, dtype=float)
        else:
            raise NotImplementedError("Model isn't trained.")


class Freq_Model:
    def __init__(self, env):
        '''
        The model of the system dynamics
        '''
        self.env = env
        self.freq = np.array([np.ones((len(env.state_space), len(env.state_space))) for i in env.control_space])
        # self.model = deepcopy(self.freq)
        self.train()
        
    def train(self, *args, **kwargs):
        self.model = deepcopy(self.freq)
        for k in self.env.control_space:
            self.model[k] = self.freq[k]/self.freq[k].sum(axis = 1)[:, np.newaxis]

            np.nan_to_num(self.model[k], copy = False)
            
    def add_data(self, s, a, ns):
        s = self._transform_s(s)
        ns = self._transform_s(ns)
        self.freq[a][s, ns] += 1
        
    def plan(self, s, acts, n_traj, det = False):
        init_s = deepcopy(s)
        s = self._transform_s(s)
        if det:
            ret = np.zeros(shape = (1, len(acts)+1, self.env.obs_dim))
            ret[:, 0, :] = init_s
            ret[0, np.arange(len(acts)+1) , self.deter_plan_one_traj(s, acts)] = 1
        else:
            ret = np.zeros(shape = (n_traj, len(acts)+1, self.env.obs_dim))
            ret[:, 0, :] = init_s
            for i in range(n_traj):
                indices = self.plan_one_traj(s, acts)
                ret[i, np.arange(len(acts)+1) , indices] = 1
                
        return np.array(ret, dtype=float)

    
    def plan_one_traj(self, s, acts):
        ret = [s]
        for a in acts:
            ns = np.random.choice(self.env.state_space, p = self.model[a][s, :])
            s = ns
            ret.append(s)
        # print(np.array(ret))
        return np.array(ret)
    
    def deter_plan_one_traj(self, s, acts):
        ret = [s]
        for a in acts:
            
            ns = self.env.state_space[self.model[a][s, :].argmax()]
            s = ns
            ret.append(s)
        # print(np.array(ret))
        return np.array(ret)
        
    def __repr__(self):
        return str(self.model)
    
    @staticmethod
    def _transform_s( s):
        
        ret = s.argmax()
        return ret
    
class True_Model_Wrapper:
    def __init__(self, real_env, stoch_plan_env, det_plan_env) -> None:
        self.env = real_env
        self.sto_env = stoch_plan_env
        self.det_env = det_plan_env
        
    
    def train(self, *args, **kwargs):
        pass
            
    def add_data(self, s, a, ns):
        pass
        
    def plan(self, s, acts, n_traj, det = False):
        init_s = deepcopy(s)
        s = self._transform_s(s)
        # print(n_traj, len(acts)+1, self.env.obs_dim)
        if det:
            ret = np.zeros(shape = (1, len(acts)+1, self.env.obs_dim))
            ret[:, 0, :] = init_s
            ret[0, np.arange(len(acts)+1) , self.deter_plan_one_traj(s, acts)] = 1
        else:
            ret = np.zeros(shape = (n_traj, len(acts)+1, self.env.obs_dim))
            ret[:, 0, :] = init_s
            for i in range(n_traj):
                indices = self.plan_one_traj(s, acts)
                ret[i, np.arange(len(acts)+1) , indices] = 1
                
        return np.array(ret, dtype=float)

    
    def plan_one_traj(self, s, acts):
        ret = [s]
        self.sto_env.reset(state = s)
        done = False
        for a in acts:
            if not done:
                ns, _, done, _ = self.sto_env.step(a)
            s = ns
            ret.append(s)
        # print('-------', np.array(ret).shape) # should be (len(acts)+1, )
        return np.array(ret)
    
    def deter_plan_one_traj(self, s, acts):
        ret = [s]
        self.det_env.reset(state = s)
        done = False
        for a in acts:
            if not done:
                ns, _, done, _ = self.det_env.step(a)
            s = ns
            ret.append(s)
        # print(np.array(ret))
        return np.array(ret)

    
    
    @staticmethod
    def _transform_s( s):
        
        ret = s.argmax()
        return ret
    


