# --------------------------------
# This module is originally from OpenAI Baselines:
# https://github.com/openai/baselines/blob/master/baselines/deepq/experiments/train_cartpole.py

# The author modified this module, to let this algorithm work with a delayed environment.
# -------------------------------


import gym

import baseline_amdp_deepq as deepq

import pathlib
ENV_PATH = str(pathlib.Path(__file__).absolute().parent.parent.parent)

import sys; sys.path.append(ENV_PATH)
from frozen_lake import Frozen_Lake_Env as Env
from utils import delayed_wrapper_amdp as delayed_wrapper
import numpy as np

from utils import initialize_uninitialized
import joblib
import argparse
from pathlib import Path




def callback(lcl, _glb):
    # stop training if reward exceeds 199
    is_solved = lcl['t'] > 100 and sum(lcl['episode_rewards'][-101:-1]) / 100 >= 199
    return is_solved

time_steps = 100000

def main(delay_steps, random_factor, wd):
    env = Env(random_factor = random_factor)
    denv = delayed_wrapper(env, 
                           delay_steps, 
                           state_dim = env.obs_dim, 
                           action_dim = env.act_dim, 
                           action_buffer= [0]*delay_steps)

    Path(wd).mkdir(parents=True, exist_ok=True) 
    
    act, debug = deepq.learn(
        denv,
        network='mlp',
        lr=1e-3,
        total_timesteps=time_steps,
        checkpoint_freq=1000,
        checkpoint_path = wd,
        checkpoint_file=f"{wd}"+"/fl{comments}_{model_name}_r"+f"{random_factor}"+"_n{time_steps}.pkl",
        learning_starts= 100,
        buffer_size=50000,
        exploration_fraction=0.1,
        exploration_final_eps=0.02,
        print_freq=10,
        # callback=callback,
        # prioritized_replay=True,
    )
    q_val_f = debug['q_values']
    print("Saving model.")
    # act.save(f"CartPole_model_r{random_factor:0.2f}_d{delay_steps}_n{time_steps}_1.pkl")

if __name__ == '__main__':
    # Example: Train a q-function when the random parameter is 0.1 and there are 5 steps of delay. 
    parser = argparse.ArgumentParser()
    parser.add_argument('--delay_steps', default=5, type=int,
                        help='The number of the delay steps.')
    parser.add_argument('--randomness_param', default=0.1, type=float,
                        help='The slippery parameter to control the randomness in transitions.')
    parser.add_argument('--work_directory', type=str,
                        help='directory to save models')
    

    args = parser.parse_args()
    main(args.delay_steps, args.randomness_param, args.work_directory)
    print(args.work_directory)
    
    # debug
    # main(5, 0.1)

